package javacomp.basics;

import java.lang.annotation.*;
import java.lang.reflect.*;

@Retention(RetentionPolicy.RUNTIME)
@Target(ElementType.TYPE)
@Inherited
@interface SpecialClass
{
    public String scope();
    public int priority();
}

@Retention(RetentionPolicy.RUNTIME)
@Target(ElementType.FIELD)
@interface SpecialField
{
    public String name();
}

@Retention(RetentionPolicy.RUNTIME)
@Target(ElementType.CONSTRUCTOR)
@interface SpecialConstructor
{
    public int priority();
}

@Retention(RetentionPolicy.RUNTIME)
@Target(ElementType.METHOD)
@Inherited
@interface SpecialMethod
{
    public String value();
}

@Retention(RetentionPolicy.RUNTIME)
@Target(ElementType.TYPE)
@Inherited
@interface SpecialBaseClass
{
    public String value() default "default";
}

@SpecialClass(scope="Class", priority=12)
class AnnotatedClass1
{
    @SpecialField(name="Field")
    private int i;

    private String s;

    @SpecialConstructor(priority = 3)
    public AnnotatedClass1(int i, String s)
    {
        this.i = i;
        this.s = s;
    }

    public AnnotatedClass1()
    {
        this.i = 0;
        this.s = "";
    }

    public int getI()
    {
        return i;
    }

    @SpecialMethod("Method")
    public String getS()
    {
        return s;
    }
}

class AnnotatedClass2 extends AnnotatedClass1
{
    private int i;
    private String s;

    public AnnotatedClass2(int i, String s)
    {
        this.i = i;
        this.s = s;
    }

    public AnnotatedClass2()
    {
        this.i = 0;
        this.s = "";
    }

    public int getI()
    {
        return i;
    }

    //@SpecialMethod("Method Overwrite")
    public String getS()
    {
        return s;
    }
}

public class AnnotationExplorer
{
    public static void main(String[] args)
    {
        if(args.length == 0)
        {
            System.out.println("arguments: list of class names");
            return;
        }
        
        for(int i = 0; i < args.length; i++)
        {
            if(i > 0)
            {
                System.out.println("-----------------------------------");
            }
            handleClass(args[i]);
        }
    }

    private static void handleClass(String className)
    {
        Class<?> classDesc = null;
        try
        {
            classDesc = Class.forName(className);
        }
        catch(ClassNotFoundException e)
        {
            System.out.println(className + " not found");
            return;
        }
        System.out.println("Class " + classDesc.getName() + ":");
        System.out.print("\tlooking for annotation SpecialClass:");
        SpecialClass s = classDesc.getAnnotation(SpecialClass.class);
        if(s != null)
        {
            System.out.println(" found, attributes:");
            System.out.println("\t\tscope=" + s.scope());
            System.out.println("\t\tpriority=" + s.priority());
        }
        else
        {
            System.out.println(" not found");
        }
        
        Field[] fields = classDesc.getDeclaredFields();
        for(Field f : fields)
        {
            System.out.println("field " + f.getName() +
                               " (" + f.getType().getName() +
                               "):");
            System.out.print("\tlooking for annotation SpecialField:");
            SpecialField sf = f.getAnnotation(SpecialField.class);
            if(sf != null)
            {
                System.out.println(" found, attributes:");
                System.out.println("\t\tname=" + sf.name());
            }
            else
            {
                System.out.println(" not found");
            }
        }
        
        Constructor<?>[] cons = classDesc.getDeclaredConstructors();
        for(Constructor<?> c : cons)
        {
            System.out.print("constructor " + c.getName() +
                             "(");
            Class<?>[] params = c.getParameterTypes();
            boolean firstParameter = true;
            for(Class<?> param : params)
            {
                if(!firstParameter)
                {
                    System.out.print(", ");
                }
                System.out.print(param.getName());
                firstParameter = false;
            }
            System.out.println("):");
            System.out.print("\tlooking for annotation SpecialConstructor:");
            SpecialConstructor sc = c.getAnnotation(SpecialConstructor.class);
            if(sc != null)
            {
                System.out.println(" found, attributes:");
                System.out.println("\t\tname=" + sc.priority());
            }
            else
            {
                System.out.println(" not found");
            }
        }
        
        Method[] methods = classDesc.getDeclaredMethods();
        for(Method m : methods)
        {
            System.out.print("method " + m.getName() + "(");
            Class<?>[] params = m.getParameterTypes();
            boolean firstParameter = true;
            for(Class<?> param : params)
            {
                if(!firstParameter)
                {
                    System.out.print(", ");
                }
                System.out.print(param.getName());
                firstParameter = false;
            }
            System.out.println("):");
            System.out.print("\tlooking for annotation SpecialMethod:");
            SpecialMethod sm = m.getAnnotation(SpecialMethod.class);
            if(sm != null)
            {
                System.out.println(" found, attributes:");
                System.out.println("\t\tvalue=" + sm.value());
            }
            else
            {
                System.out.println(" not found");
            }
        }
    }
}
