package javacomp.basics;

import java.lang.annotation.*;
import java.lang.reflect.*;

public class ReflectorForAnnotations
{
    public static void main(String[] args)
    {
        if(args.length == 0)
        {
            System.out.println("arguments: list of class names");
            return;
        }
        for(String className: args)
        {
            analyze(className);
            System.out.println("==========================================");
        }
    }

    private static void analyze(String className)
    {
        try
        {
            String headline = className + ":";
            System.out.println(headline);
            printUnderlined(headline.length());

            Class<?> c = Class.forName(className);
            
            System.out.println("  class annotations:");
            Annotation[] classAnnotations = c.getAnnotations();
            for(Annotation a : classAnnotations)
            {
                printAnnotation(a);
            }
            
            Field[] declaredAttributes = c.getDeclaredFields();
            handleAttributes(declaredAttributes);
            
            Constructor<?>[] declaredConstructors = c.getDeclaredConstructors();
            handleConstructors(declaredConstructors);
            
            Method[] declaredMethods = c.getDeclaredMethods();
            handleMethods(declaredMethods);
        }
        catch(ClassNotFoundException e)
        {
            System.out.println(className + ": not found");
        }
    }
    
    private static void printUnderlined(int length)
    {
        for(int i = 0; i < length; i++)
        {
            System.out.print("-");
        }
        System.out.println();
    }
    
    private static void printAnnotation(Annotation a)
    {
        Class<? extends Annotation> annoClass = a.annotationType();
        if(annoClass.isInterface())
        {
            System.out.print("    interface ");
        }
        else
        {
            System.out.print("    class ");
        }
        System.out.println(annoClass.getName());
        System.out.println("    {");
        Method[] declaredMethods = annoClass.getDeclaredMethods();
        for(Method m : declaredMethods)
        {
            int methodMod = m.getModifiers();
            System.out.print("      " + Modifier.toString(methodMod) + " ");
            System.out.print(m.getReturnType().getName() + " ");
            System.out.print(m.getName() + "(): ");
            try
            {
                System.out.println(m.invoke(a, new Object[]{}));
            }
            catch(Exception e)
            {
            }
        }
        System.out.println("    }");
    }
    
    private static void handleAttributes(Field[] attributes)
    {
        for(Field attribute: attributes)
        {
            System.out.println("  annotations for attribute " +
                               attribute.getName() + ":");            
            Annotation[] attributeAnnotations = attribute.getAnnotations();
            for(Annotation a : attributeAnnotations)
            {
                printAnnotation(a);
            }
        }
    }
    
    private static void handleConstructors(Constructor<?>[] constructors)
    {
        int i = 0;
        for(Constructor<?> constructor: constructors)
        {
            i++;
            System.out.println("  annotations for " + i + ". constructor:");
            Annotation[] constructorAnnotations = constructor.getAnnotations();
            for(Annotation a : constructorAnnotations)
            {
                printAnnotation(a);
            }
        }
    }
    
    private static void handleMethods(Method[] methods)
    {
        for(Method method: methods)
        {
            System.out.println("  annotations for method " +
                               method.getName() + ":");
            Annotation[] methodAnnotations = method.getAnnotations();
            for(Annotation a : methodAnnotations)
            {
                printAnnotation(a);
            }
        }
    }
}
