package javacomp.basics;

import java.lang.reflect.*;
import net.sf.cglib.proxy.Enhancer;
import net.sf.cglib.proxy.MethodInterceptor;
import net.sf.cglib.proxy.MethodProxy;

interface Counter
{
    public int add(int i);
    public int sub(int i);
}

class CounterImpl implements Counter
{
    private int c;

    public int add(int i)
    {
        c += i;
        return c;
    }

    public int sub(int i)
    {
        c -= i;
        return c;
    }
}

class StaticProxy implements Counter
{
    private Counter counter;

    public StaticProxy(Counter counter)
    {
        this.counter = counter;
    }

    public int add(int i)
    {
        System.out.println("---> javacomp.basics.Counter.add(" + i + ")");
        int result = counter.add(i);
        System.out.println("<--- return: " + result);
        return result;
    }

    public int sub(int i)
    {
        System.out.println("---> javacomp.basics.Counter.sub(" + i + ")");
        int result = counter.sub(i);
        System.out.println("<--- return: " + result);
        return result;
    }
}

class StaticProxyDerived extends CounterImpl
{
    public int add(int i)
    {
        System.out.println("---> javacomp.basics.CounterImpl.add(" + i + ")");
        int result = super.add(i);
        System.out.println("<--- return: " + result);
        return result;
    }

    public int sub(int i)
    {
        System.out.println("---> javacomp.basics.CounterImpl.sub(" + i + ")");
        int result = super.sub(i);
        System.out.println("<--- return: " + result);
        return result;
    }
}

class MyInvocationHandler implements InvocationHandler
{
    private Object target;

    public MyInvocationHandler(Object target)
    {
        this.target = target;
    }

    public Object invoke(Object proxy, Method method, Object[] args)
    {
        //System.out.println(proxy);
        String argsString = "";
        if(args != null)
        {
            for(int i = 0; i < args.length; i++)
            {
                if(i > 0)
                {
                    argsString += ", ";
                }
                argsString += args[i];
            }
        }
        System.out.println("---> " + 
                           method.getDeclaringClass().getName() +
                           "." + method.getName() +
                           "(" + argsString + ")");
   
        try
        {
            Object result = method.invoke(target, args);
            if(method.getReturnType() != void.class)
            {
                System.out.println("<--- return: " + result);
            }
            else
            {
                System.out.println("<--- return");
            }
            return result;
        }
        catch(Throwable t)
        {
            System.out.println("<--- exception: " + t.getMessage());
            return null;
        }
    }
}

class MyMethodInterceptor implements MethodInterceptor
{
    public Object intercept(Object target, Method method, Object[] args,
                            MethodProxy alternativeMethod)
        throws Throwable
    {
        String argsString = "";
        if(args != null)
        {
            for(int i = 0; i < args.length; i++)
            {
                if(i > 0)
                {
                    argsString += ", ";
                }
                argsString += args[i];
            }
        }
        System.out.println("---> " + 
                           method.getDeclaringClass().getName() +
                           "." + method.getName() +
                           "(" + argsString + ")");
   
        try
        {
            Object result = alternativeMethod.invokeSuper(target, args);
            if(method.getReturnType() != void.class)
            {
                System.out.println("<--- return: " + result);
            }
            else
            {
                System.out.println("<--- return");
            }
            return result;
        }
        catch(Throwable t)
        {
            System.out.println("<--- exception: " + t.getMessage());
            return null;
        }
    }
}

public class StaticAndDynamicProxy
{
    private static void useCounter(Counter counter)
    {
        int result = counter.add(0);
        System.out.println("start counter value: " + result);
        System.out.println("adding 7 to counter ...");
        result = counter.add(7);
        System.out.println("new counter value: " + result);
        System.out.println("subtracting 6 from counter ...");
        result = counter.sub(6);
        System.out.println("new counter value: " + result);
        System.out.println("adding 8 to counter ...");
        result = counter.add(8);
        System.out.println("new counter value: " + result);
        System.out.println("subtracting 7 from counter ...");
        result = counter.sub(7);
        System.out.println("new counter value: " + result);
    }
    
    public static void main(String[] args)
    {
        System.out.println("---------------------------");
        System.out.println("use CounterImpl");
        System.out.println("---------------------------");
        CounterImpl counterImpl = new CounterImpl();
        useCounter(counterImpl);
        System.out.println();
        
        System.out.println("---------------------------");
        System.out.println("use StaticProxy");
        System.out.println("---------------------------");
        StaticProxy staticProxy = new StaticProxy(counterImpl);
        useCounter(staticProxy);
        System.out.println();
        
        System.out.println("---------------------------");
        System.out.println("use StaticProxyDerived");
        System.out.println("---------------------------");
        StaticProxyDerived staticProxyDerived = new StaticProxyDerived();
        useCounter(staticProxyDerived);
        System.out.println();

        System.out.println("---------------------------");
        System.out.println("use DynamicProxy (1)");
        System.out.println("---------------------------");
        try
        {
            Class<?>[] implInterfaces = new Class<?>[1];
            implInterfaces[0] = Counter.class;
            Class<?> dynProxyClass = Proxy.getProxyClass(ClassLoader.getSystemClassLoader(),
                                                         implInterfaces);
            Class<?>[] formalArgs = new Class<?>[1];
            formalArgs[0] = InvocationHandler.class;
            Constructor<?> constructor = dynProxyClass.getConstructor(formalArgs);
            Object[] actualArgs = new Object[1];
            actualArgs[0] = new MyInvocationHandler(counterImpl);
            Counter dynamicProxy = (Counter) constructor.newInstance(actualArgs);
            useCounter(dynamicProxy);
        }
        catch(Exception e)
        {
            System.out.println("exception: " + e.getMessage());
        }
        System.out.println();
        
        System.out.println("---------------------------");
        System.out.println("use DynamicProxy (2)");
        System.out.println("---------------------------");
        try
        {
            Counter dynamicProxy = (Counter) Proxy.newProxyInstance(
                                                   ClassLoader.getSystemClassLoader(),
                                                   new Class<?>[] {Counter.class},
                                                   new MyInvocationHandler(counterImpl));
            useCounter(dynamicProxy);
        }
        catch(Exception e)
        {
            System.out.println("exception: " + e.getMessage());
        }
        System.out.println();
        
        /*
         * ACHTUNG ACHTUNG ACHTUNG: funktioniert so nicht.
         * Die Klasse CounterImpl muss in eine eigene Datei
         * ausgelagert werden und public sein!!!!!!
         */
        System.out.println("---------------------------");
        System.out.println("use DynamicProxy with CGLIB");
        System.out.println("---------------------------");
        try
        {
            Enhancer enhancer = new Enhancer();
            enhancer.setSuperclass(CounterImpl.class);
            enhancer.setCallback(new MyMethodInterceptor());
            CounterImpl dynamicProxy = (CounterImpl) enhancer.create();
            useCounter(dynamicProxy);
        }
        catch(Exception e)
        {
            System.out.println("exception: " + e.getMessage());
        }
    }
}
